from __future__ import annotations
from typing import TYPE_CHECKING
import asyncio


from .config import load_config
from .logging import set_logging_config, getLogger
from .utils import get_path, get_args

from .mode.simple import main_simple
from .mode.exp_multiprocess import main_experimental_multiprocess

if TYPE_CHECKING:
    pass

logger = getLogger(__name__)


def main():
    path = get_path()
    config = load_config(path)
    mode = config.running.mode
    loglevel = config.running.loglevel
    logfile = config.running.logfile

    set_logging_config(loglevel, logfile=logfile)

    logger.debug(f"Loaded config from {path}")
    logger.debug(f"Logging level set to {loglevel}")
    if logfile is not None:
        logger.debug(f"Logging to file {logfile}")

    logger.debug(f"Running in mode {mode}")
    threaded_device = get_args().threaded_device
    if threaded_device is not None:
        # TODO move at the correct place when implemented
        logger.debug(f"Running only device '{threaded_device}' in a separate thread")

    if mode.startswith("experimental"):
        logger.warning(f"Using experimental mode {config.running.mode}")

    if mode == "simple":
        asyncio.run(main_simple())
    elif mode == "experimental-multiprocess":
        main_experimental_multiprocess()
    else:
        raise ValueError(f"Unknown mode {config.running.mode}")


if __name__ == "__main__":
    main()



from __future__ import annotations
from typing import TYPE_CHECKING, cast
from typing_extensions import Optional, Annotated, Literal, TypeAlias
from pathlib import Path

import tomli
from pydantic import BaseModel, AfterValidator

if TYPE_CHECKING:
    pass


ParityFullType: TypeAlias = Literal["None", "Even", "Odd", "Mark", "Space"]
ParityLimitedType: TypeAlias = Literal["N", "E", "O", "M", "S"]
ParityType: TypeAlias = ParityFullType | ParityLimitedType


ExperimentalRunningMode: TypeAlias = Literal["experimental-multiprocess"]
RunningMode: TypeAlias = Literal["simple"] | ExperimentalRunningMode
LogLevel: TypeAlias = Literal["debug", "info", "warning", "error", "critical", "notset"]


def convert_parity(value: ParityType) -> ParityLimitedType:
    if len(value) > 0:
        value = cast(ParityLimitedType, value[0])
    return cast(ParityLimitedType, value)


class RunningConfig(BaseModel):
    logfile: Optional[str] = None
    mode: RunningMode = "simple"
    loglevel: LogLevel = "info"


class MqttAuthConfig(BaseModel):
    username: str
    password: str


class MqttConfig(BaseModel):
    host: str = "localhost"
    port: int = 1883
    auth: Optional[MqttAuthConfig] = None
    notification_topic: Optional[str] = None
    startup_wait_time: int | float = 0  # seconds (to wait for mqtt broker to start)
    pending_calls_threshold: Optional[int] = None


def one_char_string(value: str) -> str:
    if len(value) != 1:
        raise ValueError("Must be a single character")
    return value


class DefaultConfig(BaseModel):
    baudrate: int = 9600


class DeviceConfig(BaseModel):
    topic: str
    port: str
    name: Optional[str] = None
    optional: Optional[bool] = True
    logfile: Optional[str] = None
    loglevel: LogLevel = "info"
    baudrate: Optional[int] = None
    bytesize: int = 8
    parity: ParityType = "None"
    stopbits: float = 1
    timeout: Optional[int] = None
    xonxoff: bool = False
    rtscts: bool = False
    write_timeout: Optional[float] = None
    dsrdtr: bool = False
    inter_byte_timeout: Optional[float] = None
    exclusive: Optional[bool] = None
    endline_char: Annotated[str, AfterValidator(one_char_string)] = "\n"
    mqtt_start: Annotated[str, AfterValidator(one_char_string)] = "@"
    mqtt_separator: Annotated[str, AfterValidator(one_char_string)] = ":"
    log_non_mqtt_message: bool = True


class TomlConfig(BaseModel):
    running: RunningConfig = RunningConfig()
    mqtt: MqttConfig = MqttConfig()
    default: DefaultConfig = DefaultConfig()
    devices: list[DeviceConfig] = []


__global_config: Optional[TomlConfig] = None


def load_config(path: str | Path) -> TomlConfig:
    global __global_config
    if __global_config is not None:
        raise ValueError("Configuration already loaded")

    with open(path, "rb") as f:
        data = tomli.load(f)

    __global_config = TomlConfig(**data)
    return __global_config


def get_config() -> TomlConfig:
    global __global_config
    if __global_config is None:
        raise ValueError("Configuration not loaded")
    return __global_config



from __future__ import annotations
from typing import TYPE_CHECKING
from typing_extensions import Any
import multiprocessing
import asyncio
import os

from aiomqtt import Client as AioMqttClient
from ki2_python_utils import run_parallel
from mqtty import serial_device_factory, manager_setup, connect_aio_mqtt, MqttyDevice

from mqtty_cli.config import get_config, convert_parity
from mqtty_cli.logging import getLogger, configure_device_logger
from mqtty_cli.sync_to_async import (
    ask_publish_sync,
    handle_queue,
    change_mqtt_connexion_status,
)
from mqtty_cli.cli import get_args

if TYPE_CHECKING:
    pass

logger = getLogger(__name__)


def publish_notif(payload: dict[str, Any]):
    notification_topic = get_config().mqtt.notification_topic
    if notification_topic is None:
        return
    ask_publish_sync(notification_topic, payload)


def on_serial_error(device: MqttyDevice, error: Exception):
    logger.error(f"Error occured on device {device.name}: {error}")

    publish_notif(
        {
            "type": "exception",
            "source": "serial",
            "device": device.name,
            "message": str(error),
        }
    )


async def main_device(id: int):
    config = get_config()

    mqtt_config: dict[str, Any] = {
        "hostname": config.mqtt.host,
        "port": config.mqtt.port,
    }

    if config.mqtt.auth is not None:
        mqtt_config["username"] = config.mqtt.auth.username
        mqtt_config["password"] = config.mqtt.auth.password

    if config.mqtt.auth is not None:
        mqtt_config["username"] = config.mqtt.auth.username
        mqtt_config["password"] = config.mqtt.auth.password

    if config.mqtt.startup_wait_time > 0:
        await asyncio.sleep(config.mqtt.startup_wait_time)

    mqtt_client = AioMqttClient(**mqtt_config)

    if (
        isinstance(config.mqtt.pending_calls_threshold, int)
        and config.mqtt.pending_calls_threshold > 0
    ):
        mqtt_client.pending_calls_threshold = config.mqtt.pending_calls_threshold

    manager = manager_setup("async")
    manager.on_serial_error = on_serial_error

    device = config.devices[id]

    topic = device.topic
    optional = device.optional
    serial_config = device.model_dump(
        exclude={"topic", "optional", "logfile", "loglevel"}
    )
    if optional:
        if not os.path.exists(device.port):
            msg = f"Optional device '{device.port}' not found - skipping"
            logger.warning(msg)
            publish_notif(
                {
                    "type": "info",
                    "source": "setup",
                    "message": msg,
                    "device": device.name,
                    "status": "not-found",
                }
            )
            return
    if serial_config["baudrate"] is None:
        serial_config["baudrate"] = config.default.baudrate
    serial_config["parity"] = convert_parity(serial_config["parity"])
    serial_device = serial_device_factory(**serial_config)
    manager.register(topic, serial_device)
    msg = f"New device '{device.port}' on topic '{topic}'"
    logger.info(msg)
    publish_notif(
        {
            "type": "info",
            "source": "setup",
            "message": msg,
            "device": device.name,
            "status": "connected",
        }
    )

    mqtt_loop = connect_aio_mqtt(
        mqtt_client, manager=manager, conn_cb=change_mqtt_connexion_status
    )

    async def _handle_queue():
        await handle_queue(mqtt_client)

    await run_parallel(manager.loop, mqtt_loop, _handle_queue)


def run_device_process(device_id: int):
    """
    Function executed by each process to handle a specific device.

    Args:
        device_id (int): The ID of the device to be managed by this process.
    """
    config = get_config()
    device = config.devices[device_id]
    configure_device_logger(device)
    try:
        logger.info(f"Process started for device ID {device_id}")
        asyncio.run(main_device(device_id))
    except Exception as e:
        logger.error(f"Error in process for device ID {device_id}: {e}")


def main_experimental_multiprocess():
    """
    Main entry point for the experimental multiprocessing mode.
    Depending on the command-line arguments, it will either:
    - Run a single device process if a specific device ID is provided.
    - Spawn separate processes for all devices if no specific device ID is provided.
    """
    config = get_config()
    args = get_args()

    if args.threaded_device is not None:
        # Launch a single device (existing behavior)
        id = args.threaded_device
        if id < 0 or id >= len(config.devices):
            logger.error(f"Invalid device ID {id}")
            return
        asyncio.run(main_device(id))
        return

    # Launch all devices in separate processes
    logger.info("Launching all devices in separate processes...")
    processes: list[multiprocessing.Process] = []

    for id in range(len(config.devices)):
        p = multiprocessing.Process(target=run_device_process, args=(id,))
        processes.append(p)
        p.start()
        logger.info(f"Launched process for device ID {id}")

    # Wait for all processes to finish
    for p in processes:
        p.join()



from __future__ import annotations
from typing import TYPE_CHECKING
from typing_extensions import Optional
import os
import argparse
from dataclasses import dataclass
from pathlib import Path

from .logging import getLogger

if TYPE_CHECKING:
    pass

logger = getLogger(__name__)


@dataclass
class CliArgs:
    config: Optional[str] = None
    threaded_device: Optional[int] = None


__parsed_args: CliArgs | None = None


def get_args():
    global __parsed_args

    if __parsed_args is not None:
        return __parsed_args

    parser = argparse.ArgumentParser(
        description="Command-line utility designed to facilitate the interconnection"
        " between serial devices and an MQTT server. This project is based on the"
        " `mqtty` library and allows simple configuration using a TOML file."
    )
    parser.add_argument("config", nargs="?", help="Path to config file")
    parser.add_argument(
        "--threaded-device",
        type=int,
        metavar="ID",
        help="Run device with given ID in a separate thread",
    )

    __parsed_args = CliArgs(**vars(parser.parse_args()))
    return __parsed_args


def get_path():
    configpath = get_args().config
    if configpath is not None:
        return Path(configpath)
    env_conf = os.getenv("MQTTY_CONFIG")
    if env_conf is not None:
        return Path(env_conf)
    return Path("settings.toml")



from __future__ import annotations
from typing import TYPE_CHECKING
from typing_extensions import TypedDict, Any
import asyncio
import json

from aiomqtt import Client as AioMqttClient

if TYPE_CHECKING:
    pass


class QueueItem(TypedDict):
    topic: str
    payload: str


serial_error_queue: asyncio.Queue[QueueItem] = asyncio.Queue()

mqtt_connexion_status: bool = False


def ask_publish_sync(topic: str, payload: str | dict[str, Any]) -> None:
    global serial_error_queue
    if isinstance(payload, dict):
        payload = json.dumps(payload)

    serial_error_queue.put_nowait({"topic": topic, "payload": payload})


async def handle_queue(client: AioMqttClient):
    global mqtt_connexion_status
    while not mqtt_connexion_status:
        await asyncio.sleep(0.1)
    while True:
        item = await serial_error_queue.get()
        topic = item["topic"]
        payload = item["payload"]
        await client.publish(topic, payload)


def change_mqtt_connexion_status(status: bool):
    global mqtt_connexion_status
    mqtt_connexion_status = status



from __future__ import annotations
from typing import TYPE_CHECKING
from typing_extensions import Optional
import logging
from pathlib import Path

from logging import getLogger as getLogger, Logger

if TYPE_CHECKING:
    from .config import LogLevel, DeviceConfig


def get_level(level: LogLevel | str | int = "info") -> int:
    if isinstance(level, int):
        return level
    if level == "debug":
        return logging.DEBUG
    elif level == "info":
        return logging.INFO
    elif level == "warning":
        return logging.WARNING
    elif level == "error":
        return logging.ERROR
    elif level == "critical":
        return logging.CRITICAL
    elif level == "notset":
        return logging.NOTSET
    else:
        raise ValueError(f"Unknown log level: {level}")


def set_logging_config(
    level: int | str = logging.INFO, *, logfile: Path | str | None = None
):
    logging.basicConfig(
        level=get_level(level),
        format="%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    if logfile is not None:
        logger = getLogger()
        fh = logging.FileHandler(logfile)
        fh.setLevel(logging.INFO)
        formatter = logging.Formatter(
            "%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        fh.setFormatter(formatter)
        logger.addHandler(fh)


def get_devie_formatter(device: DeviceConfig):
    device_name = device.name or device.port

    format = [
        "%(asctime)s.%(msecs)03d",
        "%(name)s",
        "%(levelname)s",
        device_name,
        "%(message)s",
    ]

    formatter = logging.Formatter(
        " - ".join(format),
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    return formatter


def configure_device_logger(device: DeviceConfig, logger: Optional[Logger] = None):
    logfile = device.logfile
    level = device.loglevel
    device_name = device.name or device.port

    format = [
        "%(asctime)s.%(msecs)03d",
        "%(name)s",
        "%(levelname)s",
        device_name,
        "%(message)s",
    ]

    formatter = logging.Formatter(
        " - ".join(format),
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    if logger is None:
        logger = getLogger()

    formatter = get_devie_formatter(device)

    for handler in logger.handlers:
        handler.setFormatter(formatter)

    if logfile is not None:
        fh = logging.FileHandler(logfile)
        fh.setLevel(get_level(level))
        fh.setFormatter(formatter)
        logger.addHandler(fh)


def get_device_logger(device_name: str):
    return logging.getLogger(device_name)



from __future__ import annotations
from typing import TYPE_CHECKING
from typing_extensions import Any
import asyncio
import os

from aiomqtt import Client as AioMqttClient
from ki2_python_utils import run_parallel
from mqtty import serial_device_factory, manager_setup, connect_aio_mqtt, MqttyDevice

from mqtty_cli.config import get_config, convert_parity
from mqtty_cli.logging import getLogger
from mqtty_cli.sync_to_async import (
    ask_publish_sync,
    handle_queue,
    change_mqtt_connexion_status,
)

if TYPE_CHECKING:
    pass

logger = getLogger(__name__)


def publish_notif(payload: dict[str, Any]):
    notification_topic = get_config().mqtt.notification_topic
    if notification_topic is None:
        return
    ask_publish_sync(notification_topic, payload)


def on_serial_error(device: MqttyDevice, error: Exception):
    logger.error(f"Error occured on device {device.name}: {error}")

    publish_notif(
        {
            "type": "exception",
            "source": "serial",
            "device": device.name,
            "message": str(error),
        }
    )


async def main_simple():
    config = get_config()

    if len(config.devices) == 0:
        logger.warning("No devices configured")
        return

    mqtt_config: dict[str, Any] = {
        "hostname": config.mqtt.host,
        "port": config.mqtt.port,
    }

    if config.mqtt.auth is not None:
        mqtt_config["username"] = config.mqtt.auth.username
        mqtt_config["password"] = config.mqtt.auth.password

    if config.mqtt.auth is not None:
        mqtt_config["username"] = config.mqtt.auth.username
        mqtt_config["password"] = config.mqtt.auth.password

    if config.mqtt.startup_wait_time > 0:
        await asyncio.sleep(config.mqtt.startup_wait_time)

    mqtt_client = AioMqttClient(**mqtt_config)

    if (
        isinstance(config.mqtt.pending_calls_threshold, int)
        and config.mqtt.pending_calls_threshold > 0
    ):
        mqtt_client.pending_calls_threshold = config.mqtt.pending_calls_threshold

    manager = manager_setup("async")
    manager.on_serial_error = on_serial_error

    for device in config.devices:
        topic = device.topic
        optional = device.optional
        serial_config = device.model_dump(
            exclude={"topic", "optional", "logfile", "loglevel"}
        )
        if optional:
            if not os.path.exists(device.port):
                msg = f"Optional device '{device.port}' not found - skipping"
                logger.warning(msg)
                publish_notif(
                    {
                        "type": "info",
                        "source": "setup",
                        "message": msg,
                        "device": device.name,
                        "status": "not-found",
                    }
                )
                continue
        if serial_config["baudrate"] is None:
            serial_config["baudrate"] = config.default.baudrate
        serial_config["parity"] = convert_parity(serial_config["parity"])
        serial_device = serial_device_factory(**serial_config)
        manager.register(topic, serial_device)
        msg = f"New device '{device.port}' on topic '{topic}'"
        logger.info(msg)
        publish_notif(
            {
                "type": "info",
                "source": "setup",
                "message": msg,
                "device": device.name,
                "status": "connected",
            }
        )

    mqtt_loop = connect_aio_mqtt(
        mqtt_client, manager=manager, conn_cb=change_mqtt_connexion_status
    )

    async def _handle_queue():
        await handle_queue(mqtt_client)

    await run_parallel(manager.loop, mqtt_loop, _handle_queue)
